# Interface implementation Ball And Beam ====================================================

# The ball and beam can be used in two modes, either just the Beam, in which case there is a
# single output (measurement signal) or the BallAndBeam, in which case there are two.
# There are a few union types defined for convenience, these are
# AbstractBeam = Union{Beam, BeamSimulator}
# AbstractBallAndBeam   = Union{BallAndBeam, BallAndBeamSimulator}
# AbstractBeamOrBallAndBeam = All types
# Although not Abstract per se, the names AbstractBeam etc. were chosen since this reflects
# their usage in dispatch.

export Beam, BeamSimulator, AbstractBeam, BallAndBeam, BallAndBeamSimulator, AbstractBeamOrBallAndBeam

@with_kw struct Beam <: PhysicalProcess
    h::Float64 = 0.01
    bias::Float64 = 0.0
    stream::LabStream = ComediStream()
    measure::AnalogInput10V = AnalogInput10V(0)
    control::AnalogOutput10V = AnalogOutput10V(1)
end

include("define_beam_system.jl")
const beam_system, nice_beam_controller = define_beam_system()
# nice_beam_controller gives ϕₘ=56°, Aₘ=4, Mₛ = 1.6. Don't forget to discretize it before use
struct BeamSimulator <: SimulatedProcess
    h::Float64
    s::SysFilter
    BeamSimulator() = new(0.01, SysFilter(beam_system, 0.01))
    BeamSimulator(h::Real) = new(Float64(h), SysFilter(beam_system, h))
end

@with_kw struct BallAndBeam <: PhysicalProcess
    h::Float64 = 0.01
    bias::Float64 = 0.0
    stream::LabStream = ComediStream()
    measure1::AnalogInput10V = AnalogInput10V(0)
    measure2::AnalogInput10V = AnalogInput10V(1)
    control::AnalogOutput10V = AnalogOutput10V(0)
end

struct BallAndBeamSimulator <: SimulatedProcess
    h::Float64
    s::SysFilter
end

const AbstractBeam              = Union{Beam, BeamSimulator}
const AbstractBallAndBeam       = Union{BallAndBeam, BallAndBeamSimulator}
const AbstractBeamOrBallAndBeam = Union{AbstractBeam, AbstractBallAndBeam}

num_outputs(p::AbstractBeam)             = 1
num_outputs(p::AbstractBallAndBeam)      = 2
num_inputs(p::AbstractBeamOrBallAndBeam) = 1
outputrange(p::AbstractBeam)             = [(-10,10)]
outputrange(p::AbstractBallAndBeam)      = [(-10,10),(-1,1)] # Beam angle, Ball position
inputrange(p::AbstractBeamOrBallAndBeam) = [(-10,10)]
isstable(p::AbstractBeam)                = true
isstable(p::AbstractBallAndBeam)         = false
isasstable(p::AbstractBeamOrBallAndBeam) = false
sampletime(p::AbstractBeamOrBallAndBeam) = p.h
bias(p::AbstractBeamOrBallAndBeam)       = p.bias
bias(p::BeamSimulator)                   = 0
bias(p::BallAndBeamSimulator)            = 0


control(p::AbstractBeamOrBallAndBeam, u) = send(p.control,u)
control(p::BeamSimulator, u)             = p.s(u)
control(p::BallAndBeamSimulator, u)      = error("Not yet implemented")

measure(p::Beam)                         = read(p.measure)
measure(p::BallAndBeam)                  = [read(p.measure1), read(p.measure2)]
measure(p::BeamSimulator)                = vecdot(p.s.sys.C,p.s.state)
measure(p::BallAndBeamSimulator)         = error("Not yet implemented")


initialize(p::Beam)                    = init_devices!(p.stream, p.measure, p.control)
initialize(p::BallAndBeam)             = init_devices!(p.stream, p.measure1,
                                                       p.measure2, p.control)
finalize(p::AbstractBeamOrBallAndBeam) = close(p.stream)
initialize(p::BallAndBeamSimulator)    = nothing
finalize(p::BallAndBeamSimulator)      = nothing
initialize(p::BeamSimulator)           = p.s.state .*= 0
finalize(p::BeamSimulator)             = nothing
